{********************************************************************}
{* Common functions and procedures unit for LogicGates6 application *}
{********************************************************************}

unit common;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, StdCtrls, ExtCtrls, Graphics, Dialogs;

// Public types for switches and LEDs description

type
  // Switches description
  TSwitch = record
    SwitchShapeA, SwitchShapeB: TShape;                                        // horizontal and vertical parts (shapes) of the switch
    SwitchShapeAOffPos: Integer;                                               // vertical position ("Top") of the horiz. switch part, when switch is off
    SwitchOn: Boolean;                                                         // switch on (True) or off (False)
  end;
  // LEDs description
  TLed = record
    LedShape: TShape;                                                          // shape of the LED
    LedOn: Boolean;                                                            // LED on (True) or off (False)
  end;

// Public constants

const
  // "Virtual switch" array (cf. SwitchOnOff procedure)
  VirtualSwitches: array of TSwitch = nil;
  // Number subscripts UTF-8 codes
  SUB_Numbers: array[0..9] of string = (
    #$E2#$82#$80, #$E2#$82#$81, #$E2#$82#$82, #$E2#$82#$83, #$E2#$82#$84,
    #$E2#$82#$85, #$E2#$82#$86, #$E2#$82#$87, #$E2#$82#$88, #$E2#$82#$89
  );

// Public procedures for combinational circuits simulation (called from the different simulation windows)

procedure LedOn(var Led: TLed);
procedure LedOff(var Led: TLed);
procedure LedOnOff(TurnLedOn: Boolean; var Led: TLed);
procedure LedColor(Colour: TColor; var Led: TLed);
procedure SetLedColor(L: Integer; DefaultColor: TColor; Colored: Boolean; var Led: TLed);
procedure SetLedColors(DefaultColor: TColor; Colored: Boolean; var Leds: array of TLed);
procedure SwitchOn(var Switch: TSwitch);
procedure SwitchOff(var Switch: TSwitch);
procedure SwitchToggle(var Switch: TSwitch);
procedure SwitchOnOff(Switch: Integer; IsRealSwitch, PushSwitchOn: Boolean; var Switches: array of TSwitch; var Leds: array of TLed;
  var Input: array of Boolean; SwitchesOff, NoWarning: Boolean);
procedure Encoder42(var Input, Output: array of Boolean; var Leds: array of TLed);
procedure Encoder83(var Input, Output: array of Boolean; var Leds: array of TLed);
procedure Encoder42Pri(var Input, Output: array of Boolean; var Leds: array of TLed; out Validity: Boolean; var LedValidity: TLed);
procedure Decoder24(Enable: Boolean; var Input, Output: array of Boolean; Leds: array of TLed);
procedure Multiplexer41(var Selections, Inputs: array of Boolean; var Line: Integer; var Output: Boolean;
  var LedsOutput: array of TLed; var LedOutput: TLed);
procedure DeMultiplexer14(var Selections: array of Boolean; Input: Boolean; var Line: Integer;
  var Outputs: array of Boolean; var LedsOutput: array of TLed);

implementation

// Switches related constants
const
  SwitchShapeBOffPos = -15;                                                    // offset of vertical position ("Top") of the vert. switch part
  SwitchShapeOnPos = +12;                                                      // offset of vertical position ("Top") of the switch parts, when switch is on

{ Push a given switch on }

procedure SwitchOn(var Switch: TSwitch);

begin
  Switch.SwitchOn := True;
  Switch.SwitchShapeA.Top := Switch.SwitchShapeAOffPos + SwitchShapeOnPos;     // vertical position of horizontal switch part
  Switch.SwitchShapeB.Top := Switch.SwitchShapeA.Top + SwitchShapeBOffPos;     // vertical position of vertical switch part
end;

{ Push a given switch off }

procedure SwitchOff(var Switch: TSwitch);

begin
  Switch.SwitchOn := False;
  Switch.SwitchShapeA.Top := Switch.SwitchShapeAOffPos;                        // vertical position of horizontal switch part
  Switch.SwitchShapeB.Top := Switch.SwitchShapeA.Top + SwitchShapeBOffPos;     // vertical position of vertical switch part
end;

{ Push a given switch on, if it is off and vice-versa }

procedure SwitchToggle(var Switch: TSwitch);

begin
  if Switch.SwitchOn then
    SwitchOff(Switch)
  else
    SwitchOn(Switch);
end;

{ Turn a given LED on }

procedure LedOn(var Led: TLed);

begin
  Led.LedOn := True;
  Led.LedShape.Brush.Style := bsSolid;                                         // LED on is shown by solid brush style (fully colored shaped)
end;

{ Turn a given LED off }

procedure LedOff(var Led: TLed);

begin
  Led.LedOn := False;
  Led.LedShape.Brush.Style := bsDiagCross;                                     // LED on is shown by diagonal cross brush style (sparsely colored shaped)
end;

{ Turn LED on or off, depending on given logic value }

procedure LedOnOff(TurnLedOn: Boolean; var Led: TLed);

begin
  if TurnLedOn then
    LedOn(Led)
  else
    LedOff(Led);
end;

{ Set LED color to a given value }

procedure LedColor(Colour: TColor; var Led: TLed);

begin
  Led.LedShape.Brush.Color := Colour;
end;

{ Set color of a given LED (array element) to the corr. value (corr. array element) OR to given default value }

procedure SetLedColor(L: Integer; DefaultColor: TColor; Colored: Boolean; var Led: TLed);

const
  // Values to be used for colored LEDs (1-1 correspondance with LED array)
  ColoredLedsColors: array[0..3] of TColor = (
    $00FACE87, $00FFBF00, $00FF901E, clBlue
  );

begin
  if Colored then
    // LEds are colored LEDs
    LedColor(ColoredLedsColors[L], Led)
  else
    // LEds aren't colored LEds
    LedColor(DefaultColor, Led);
end;

{ Set color of all LEDs (array elements) to the corr. values (corr. array elements) OR to given default value }

procedure SetLedColors(DefaultColor: TColor; Colored: Boolean; var Leds: array of TLed);

var
  I: Integer;

begin
  for I := 0 to Length(Leds) - 1 do
    SetLedColor(I, DefaultColor, Colored, Leds[I])
end;

{ "Special" switch on/off procedure }

procedure SwitchOnOff(Switch: Integer; IsRealSwitch, PushSwitchOn: Boolean; var Switches: array of TSwitch; var Leds: array of TLed;
  var Input: array of Boolean; SwitchesOff, NoWarning: Boolean);

// Push on or off a given switch (switches array element) depending on Boolean variable PushSwitchOn
// Turn corr. LEd (LEDs array element) on or off
// Set corresponding input line (Booleans array element) to High or Low
// Push or not all switches off, beforing pushing the switch
// Display or not a warning or no switch is on or if more than one switch is on

// The procedure may also be used with "virtual" switches (I used these with the multiplexer simulation, where the input lines are
// set High or Low by clicking on the LED, thus no physical switch there). In this case, call the procedure with Switches = VirtualSwitches
// and set IsRealSwitch to False.

var
  N, I: Integer;

begin
  if SwitchesOff then begin
    // Turn all switches off (used in the encoders simulations in order to only have valid inputs)
    for I := 0 to Length(Switches) - 1 do begin
      if IsRealSwitch then
        SwitchOff(Switches[I]);
      LedOff(Leds[I]);
      Input[I] := False;
    end;
  end;
  if PushSwitchOn then begin
    // Switch has to be pushed on
    if IsRealSwitch then                                                       // if there is a physical switch, push it on
      SwitchOn(Switches[Switch]);
    LedOn(Leds[Switch]);                                                       // turn the corr. input LED on
    Input[Switch] := True;                                                     // set the corr. input line to High
  end
  else begin
    // Switch has to be pushed off
    if IsRealSwitch then                                                       // if there is a physical switch, push it off
      SwitchOff(Switches[Switch]);
    LedOff(Leds[Switch]);                                                      // turn the corr. input LED off
    Input[Switch] := False;                                                    // set the corr. input line to Low
  end;
  // Check if exactly one switch is pushed on. This allows to issue a warning in the case of an invalid encoder input
  // (for all other simulations, NoWarnings has to be set to True!)
  if not NoWarning then begin
    N := 0;
    for I := 0 to Length(Switches) - 1 do begin
      if Input[I] then
        Inc(N);
    end;
    if N = 0 then
      MessageDlg('Encoder issue', 'All input lines shouldn''t be Low at the same time!', mtWarning, [mbOK], 0)
    else if N > 1 then
      MessageDlg('Encoder issue', 'Only one single input line should be active High at a time!', mtWarning, [mbOK], 0);
  end;
end;

{ ---------------------------------------------------------------------------------------------------- }
{                              Combinational circuits  simulaion routines                              }
{ ---------------------------------------------------------------------------------------------------- }
{ The code is nothing else than the implementation of the logic gates as Free Pascal Boolean functions }
{ The procedure sets the output line(s) status (Boolean variable) and turns the corr. LED on or off    }
{ ---------------------------------------------------------------------------------------------------- }

{ 4-to-2 encoder simulation }

procedure Encoder42(var Input, Output: array of Boolean; var Leds: array of TLed);

var
  I: Integer;

begin
  Output[1] := Input[3] or Input[2];
  Output[0] := Input[3] or Input[1];
  for I := 0 to 1 do
    LedOnOff(Output[I], Leds[I]);
end;

{ 8-to-3 encoder simulation }

procedure Encoder83(var Input, Output: array of Boolean; var Leds: array of TLed);

var
  I: Integer;

begin
  Output[2] := Input[7] or Input[6] or Input[5] or Input[4];
  Output[1] := Input[7] or Input[6] or Input[3] or Input[2];
  Output[0] := Input[7] or Input[5] or Input[3] or Input[1];
  for I := 0 to 2 do
    LedOnOff(Output[I], Leds[I]);
end;

{ 4-to-2 priority encoder simulation }

procedure Encoder42Pri(var Input, Output: array of Boolean; var Leds: array of TLed; out Validity: Boolean; var LedValidity: TLed);

var
  I: Integer;

begin
  Output[1] := Input[3] or Input[2];
  Output[0] := Input[3] or (not Input[2] and Input[1]);
  Validity := Input[3] or Input[2] or Input[1] or Input[0];
  for I := 0 to 1 do
    LedOnOff(Output[I], Leds[I]);
  LedOnOff(Validity, LedValidity);
end;

{ 2-to-4 decoder simulation }

procedure Decoder24(Enable: Boolean; var Input, Output: array of Boolean; Leds: array of TLed);

var
  I: Integer;

begin
  Output[3] := Enable and Input[1] and Input[0];
  Output[2] := Enable and Input[1] and (not Input[0]);
  Output[1] := Enable and (not Input[1]) and Input[0];
  Output[0] := Enable and (not Input[1]) and (not Input[0]);
  for I := 0 to 3 do
    LedOnOff(Output[I], Leds[I]);
end;

{ 1x4 multiplexer simulation }

procedure Multiplexer41(var Selections, Inputs: array of Boolean; var Line: Integer; var Output: Boolean;
  var LedsOutput: array of TLed; var LedOutput: TLed);

// This procedure also returns the input line selected (integer value: 0..3) and turns the AND gates output LEDs on or off

var
  I: Integer;
  Outputs: array of Boolean;

begin
  SetLength(Outputs, Length(Inputs));
  Outputs[3] := Inputs[3] and Selections[1] and Selections[0];
  Outputs[2] := Inputs[2] and Selections[1] and (not Selections[0]);
  Outputs[1] := Inputs[1] and (not Selections[1]) and Selections[0];
  Outputs[0] := Inputs[0] and (not Selections[1]) and (not Selections[0]);
  Output := Outputs[3] or Outputs[2] or Outputs[1] or Outputs[0];
  Line := 0;
  if Selections[1] then
    Line += 2;
  if Selections[0] then
    Line += 1;
  for I := 0 to 3 do
    LedOnOff(Outputs[I], LedsOutput[I]);
  LedOnOff(Output, LedOutput);
end;

{ 4x1 demultiplexer simulation }

procedure DeMultiplexer14(var Selections: array of Boolean; Input: Boolean; var Line: Integer;
  var Outputs: array of Boolean; var LedsOutput: array of TLed);

// This procedure also returns the output line selected (integer value: 0..3)

var
  I: Integer;

begin
  Outputs[3] := Input and Selections[1] and Selections[0];
  Outputs[2] := Input and Selections[1] and (not Selections[0]);
  Outputs[1] := Input and (not Selections[1]) and Selections[0];
  Outputs[0] := Input and (not Selections[1]) and (not Selections[0]);
  Line := 0;
  if Selections[1] then
    Line += 2;
  if Selections[0] then
    Line += 1;
  for I := 0 to 3 do
    LedOnOff(Outputs[I], LedsOutput[I]);
end;

end.

